;************************************************************************************************
;*                                  Autocom.ASM                                                 *
;*   :                                                       *
;*              : Copyright (c) 2013 www.v-50m.narod.ru                                         *
;*              : All Rights Reserved                                                           *
;*          : 05.09.2013                                                                    *
;*        : 1.1                                                                           *
;*      :                                            *
;*                                                                 *
;*                       PIC16F628                         *                                        
;************************************************************************************************

#DEFINE PIC16F628

#INCLUDE <p16f628a.inc>


;  ----------------------------------------------------------------------------------

    __CONFIG    _CP_OFF & _WDT_ON & _PWRTE_ON & _XT_OSC & _LVP_OFF & _MCLRE_ON & _BODEN_OFF & _DATA_CP_OFF
    __IDLOCS    0xAC11

;---------------------------   -------------------------
#DEFINE		P_SIGNAL	PORTB , 00	;     
#DEFINE		OUT_SIGNAL	PORTB , 01	;   
#DEFINE		INVERT_SIGNAL	PORTB , 02	;  
#DEFINE		N_SIGNAL	PORTB , 03	;     
#DEFINE		MAP		PORTB , 04	;  
#DEFINE		LOCK_ON		PORTB , 05	;  - ""

#DEFINE		IN_LED		PORTA , 02	;  ""
#DEFINE		LOCK_LED	PORTA , 03	;  ""

;---------------------------  ----------------------------------------------------------
CBLOCK	20h
W_temp
Status_temp
Ignition
DelayCounter : 3
ENDC

CBLOCK 70h
X				
Y
T : 2			
Temp : 4
Tmin : 2
Tmax : 2
Tnp : 2			
ENDC
	
#DEFINE		LowRPM		Ignition , 00
#DEFINE		Spark		Ignition , 01
#DEFINE		Break		Ignition , 02
#DEFINE		SkipPulse	Ignition , 03
#DEFINE		CalibrationEnable		Ignition , 04

;---------------------------  ----------------------------------------------------------
PORTA_RESET		EQU	B'00000'
PORTB_RESET		EQU	B'00000000'
TRISA_RESET		EQU	B'00000'
TRISB_RESET		EQU	B'00111001'

INTCON_RESET		EQU	B'11000000'
T1CON_RESET		EQU	B'00110000'
CCP1CON_RESET		EQU	B'00001010'
;T2CON_RESET		EQU	B'00000000'
;PR2_RESET		EQU	.100	;x+9 
PIE1_RESET		EQU	B'00000110'
OPTION_RESET		EQU	B'10000000'

IGNITION_RESET		EQU	B'00001001'
;INPUT_DELAY		EQU	.10	;27+3x

#INCLUDE <IgnTable.asm>

;================================================================================================
; ----------------------------------------  ----------------------------------------------
;================================================================================================   
MMAC	MACRO	A , BIT , U2 , U1			;    MULTIPLY_16x8
	BTFSC	A , BIT
	ADDWF	U2 , f
	RRF	U2 , f
	RRF	U1 , f
	ENDM

multiply16x8	MACRO	AH , AL , BL , R3 , R2 , R1	;   16    
	CLRF	R3					; 8 . AH:AL - , 
	CLRF	R1					; BL -  R3:R2:R1 - 					
	CLRC
	MOVFW	BL
	MMAC	AL , 0 , R3 , R1
	MMAC	AL , 1 , R3 , R1
	MMAC	AL , 2 , R3 , R1
	MMAC	AL , 3 , R3 , R1
	MMAC	AL , 4 , R3 , R1
	MMAC	AL , 5 , R3 , R1
	MMAC	AL , 6 , R3 , R1
	MMAC	AL , 7 , R3 , R1
	CLRF	R2
	MMAC	AH , 0 , R3 , R2
	MMAC	AH , 1 , R3 , R2
	MMAC	AH , 2 , R3 , R2
	MMAC	AH , 3 , R3 , R2
	MMAC	AH , 4 , R3 , R2
	MMAC	AH , 5 , R3 , R2
	MMAC	AH , 6 , R3 , R2
	MMAC	AH , 7 , R3 , R2
	ENDM	

divBy512	MACRO	TMPCNTR , X2 , X1 , X0
	MOVLW	0xF7
	MOVWF	TMPCNTR

NXT_DIV BCF	STATUS, C
	RRF	X2 , f
	RRF	X1 , f
	RRF	X0 , f

	INCFSZ	TMPCNTR , f
	GOTO NXT_DIV
	ENDM


;================================================================================================
; ---------------------------------------   -----------------------------------------
;================================================================================================   
     ORG 0000h
	goto Start

;================================================================================================
; ---------------------------------------   -------------------------------------
;================================================================================================
    ORG	0004h
	btfsc	PIR1, TMR2IF
	goto	TMR2_interrupt

	btfsc	PIR1, CCP1IF	
	goto	CCP1_Interrupt

	goto $		;     ,    
			; .    WDT
;================================================================================================
; ---------------------------------------   TMR2 ------------------------------------
;================================================================================================
TMR2_interrupt
	bcf	OUT_SIGNAL
	movwf	W_temp
	swapf	STATUS, w
	movwf	Status_temp
	bcf	T2CON , TMR2ON
	clrf	TMR2
	bcf	PIR1 , TMR2IF
	swapf	Status_temp, w
	movwf	STATUS
	swapf	W_temp, f
	swapf	W_temp, w
	retfie

;================================================================================================
; ---------------------------------------   CCP1 ------------------------------------
;================================================================================================
CCP1_Interrupt
	bcf	PIR1, CCP1IF
	btfss	Break
	btfsc	LowRPM
	retfie
	btfsc	Spark
	retfie

	bsf	OUT_SIGNAL
	bsf	T2CON, TMR2ON
;	bcf	INVERT_SIGNAL
	bsf	Spark
	retfie	

;================================================================================================
; -----------------        ----------------------------
;================================================================================================
Start	
	clrwdt
	movlw	PORTA_RESET	;     
	movwf	PORTA		;
	movlw	PORTB_RESET	;
	movwf	PORTB		;

	movlw	0x07		;   
	movwf	CMCON		;

	bsf	STATUS , RP0		
	movlw	TRISA_RESET	;   
	movwf	TRISA		;
	movlw	TRISB_RESET	;
	movwf	TRISB		;
	movlw	PIE1_RESET	;
	movwf	PIE1		;
	movlw	OPTION_RESET	;
	movwf	OPTION_REG	;
	movlw	PR2_RESET	;
	movwf	PR2		;

	movlw	0x78		;    EEPROM
	movwf	EEADR		;
	bsf	EECON1, RD	;
	movf	EEDATA, w	;
	movwf	Tmax+1		;
	incf	EEADR, f	;
	bsf	EECON1, RD	;
	movf	EEDATA, w	;
	movwf	Tmax+0		;
	incf	EEADR, f	;
	bsf	EECON1, RD	;
	movf	EEDATA, w	;
	movwf	Tmin+1		;
	incf	EEADR, f	;
	bsf	EECON1, RD	;
	movf	EEDATA, w	;
	movwf	Tmin+0		;
	bcf	STATUS, RP0	

	clrf	TMR1L
	clrf	TMR1H
	movlw	0xFF
	movwf	CCPR1H	
	movwf	CCPR1L
	movlw	CCP1CON_RESET
	movwf	CCP1CON
	movlw	T1CON_RESET	;  
	movwf	T1CON		;
	movlw	T2CON_RESET	; 
	movwf	T2CON		;

	movlw	IGNITION_RESET
	movwf	Ignition

	movlw	INTCON_RESET	;  
	movwf	INTCON		;

	btfss	LOCK_ON
	bsf	LOCK_LED

;================================================================================================
; ------------------------------------   -----------------------------------
;================================================================================================
Main
	clrwdt
	btfsc	LowRPM
	call	InvertInLed
	
	btfss	INTCON, INTF
	goto	Main

	btfsc	N_SIGNAL
	goto	PP_interrupt
	goto	NP_interrupt

;================================================================================================
; ---------------------------------------   PP --------------------------------------
;================================================================================================
PP_interrupt
	call	Delay
	bcf	INTCON, INTF
	btfsc	P_SIGNAL
	goto	Main

	bcf	INTCON, GIE
	bcf	T1CON, TMR1ON
	movf	TMR1H, w	
	movwf	Tnp+1		;     Tnp 
	movf	TMR1L, w
	movwf	Tnp+0
	bsf	T1CON, TMR1ON
	bsf	INTCON, GIE

	btfss	LowRPM	
	goto	ExitPPint
	bsf	OUT_SIGNAL
	bsf	T2CON, TMR2ON
	bsf	Spark
ExitPPint
	bsf	SkipPulse
	goto	Main

;================================================================================================
; ---------------------------------------   NP --------------------------------------
;================================================================================================
NP_interrupt
	call	Delay
	bcf	INTCON, INTF
	btfsc	N_SIGNAL
	goto	Main	

#ifdef double_pulse	
	btfss	SkipPulse
	goto	Change_const
	bcf	SkipPulse
	goto	Main
#else
	nop
	nop
	nop
	nop
#endif

Change_const
	bcf	INTCON, GIE
	bcf	T1CON, TMR1ON
	movf	TMR1H, w	
	movwf	T+1		;     T ...
	movf	TMR1L, w
	movwf	T+0
	clrf	TMR1L		; ...   
	bsf	T1CON, TMR1ON
	clrf	TMR1H		;
	bsf	INTCON, GIE
	
	bcf	Break
	bcf	LowRPM
	bcf	Spark

	btfsc	PIR1, TMR1IF
	bsf	LowRPM
	bcf	PIR1, TMR1IF

CompareTmax
	movf	Tmax+0, w 	;  T  Tmax
	subwf	T+0, w		;
	movf	Tmax+1, w	;
	btfss	STATUS, C	;
	incfsz	Tmax+1, w 	;
	subwf	T+1, w		;
	btfss	STATUS, C	
	goto	$+2
	bsf	LowRPM		;  ,  T > Tmax (  )

CompareTmin
	movf	Tmin+0, w 	;  T  Tmin
	subwf	T+0, w		;
	movf	Tmin+1, w	;
	btfss	STATUS, C	;
	incfsz	Tmin+1, w 	;
	subwf	T+1, w		;

	btfss	STATUS, C
	btfsc	LOCK_ON
	goto	$+2
	bsf	Break		;  ,  T < Tmin     (  )

	btfsc	CalibrationEnable	;     ...
	btfsc	LowRPM			
	goto	$+2
	goto	Calibration		; ...     ,      

	movf	T+1, w		; W = 31 - [ (T)] 
	sublw	.31		
	btfss	STATUS, C	;  [ (T)] > 31 ...
	movlw	.31		; ... W = 31
	call	Const_Adr	;     EEPROM     T

	btfss	MAP		;     
	addlw	.60

	bsf	STATUS, RP0	;    EEPROM
	movwf	EEADR		;
	bsf	EECON1, RD	;
	movf	EEDATA, w	;
	movwf	X		;

	btfsc	STATUS, Z
	bsf	LowRPM

	incf	EEADR, f	;
	bsf	EECON1, RD	;
	movf	EEDATA, w	;
	movwf	Y		;
	bcf	STATUS, RP0	;


	multiply16x8	T+1 , T+0 , X , Temp+2 , Temp+1 , Temp+0	; Temp = T * X
	nop
	divBy512	Temp+3 , Temp+2 , Temp+1 , Temp+0		; Temp = Temp / 512

	movf	Y , w		; CCPR1 = Temp - Y
	subwf	Temp+0, w	;
	btfss	STATUS, C	;
	decf	Temp+1, f	;
	movwf	CCPR1L		;
	movf	Temp+1, w	;
	movwf	CCPR1H 		;

	btfss	LowRPM
	bsf	IN_LED

	btfsc	Break
	bcf	IN_LED

        btfss   LOCK_ON		
        goto    $+3
        bcf     LOCK_LED
        goto    Main
	bsf     LOCK_LED	;  ,    ""

	goto	Main	

;================================================================================================
; -----------------------------------------------------------------------------------------------
;================================================================================================
InvertInLed
	decfsz	DelayCounter+0, f
	return
	decfsz	DelayCounter+1, f
	return
	clrw
	btfss   IN_LED
	addlw	0x01
        btfsc   STATUS    , Z
	bcf	IN_LED
        btfss   STATUS    , Z
	bsf	IN_LED
	return

;================================================================================================
; -----------------------------------------------------------------------------------------------
;================================================================================================
InvertLockLed
	clrwdt
	decfsz	DelayCounter+0, f
	return
	decfsz	DelayCounter+1, f
	return
	clrw
	btfss   LOCK_LED
	addlw	0x01
        btfsc   STATUS    , Z
	bcf	LOCK_LED
        btfss   STATUS    , Z
	bsf	LOCK_LED
	return

;================================================================================================
; --------------------------      -------------------------------
;================================================================================================
Delay
	movlw	INPUT_DELAY
	movwf	DelayCounter+2
DelayLoop
	decfsz	DelayCounter+2, f
	goto	DelayLoop
	return

;================================================================================================
; ---------------------------------------  --------------------------------------------
;================================================================================================
Calibration
	bsf	STATUS, RP0
	bsf	EECON1, WREN
	movlw	0x7C
	movwf	EEADR
	movf	Tnp+1, w
	movwf	EEDATA

	bcf	INTCON, GIE
	movlw	0x55
	movwf	EECON2
	movlw	0xAA
	movwf	EECON2
	bsf	EECON1, WR
	
	clrwdt
	btfsc	EECON1, WR
	goto	$-2

	incf	EEADR, f
	movf	Tnp+0, w
	movwf	EEDATA

	movlw	0x55
	movwf	EECON2
	movlw	0xAA
	movwf	EECON2
	bsf	EECON1, WR

	clrwdt
	btfsc	EECON1, WR
	goto	$-2
	
	incf	EEADR, f
	movf	T+1, w
	movwf	EEDATA

	movlw	0x55
	movwf	EECON2
	movlw	0xAA
	movwf	EECON2
	bsf	EECON1, WR

	clrwdt
	btfsc	EECON1, WR
	goto	$-2

	incf	EEADR, f
	movf	T+0, w
	movwf	EEDATA

	movlw	0x55
	movwf	EECON2
	movlw	0xAA
	movwf	EECON2
	bsf	EECON1, WR

	clrwdt
	btfsc	EECON1, WR
	goto	$-2
		
	bcf	EECON1, WREN
	bcf	STATUS, RP0

	call	InvertLockLed
	goto	$-2

;================================================================================================
; --------------------------------    EEPROM -------------------------------------
;================================================================================================
	ORG	0200h	

Const_Adr
	clrf	PCLATH
	bsf	PCLATH , 01
	addwf	PCL , f		;     EEPROM     T
	retlw	.58	; 31
	retlw	.56	; 30
	retlw	.54	; 29
	retlw	.52	; 28
	retlw	.50	; 27
	retlw	.48	; 26
	retlw	.46	; 25
	retlw	.44	; 24
	retlw	.42	; 23
	retlw	.40	; 22
	retlw	.38	; 21
	retlw	.36	; 20
	retlw	.34	; 19
	retlw	.32	; 18
	retlw	.30	; 17
	retlw	.28	; 16
	retlw	.26	; 15
	retlw	.24	; 14
	retlw	.22	; 13	
	retlw	.20	; 12
	retlw	.18	; 11
	retlw	.16	; 10
	retlw	.14	; 9
	retlw	.12	; 8
	retlw	.10	; 7
	retlw	.8	; 6
	retlw	.6	; 5
	retlw	.4	; 4
	retlw	.2	; 3	
	retlw	.0	; 2
	retlw	.0	; 1
	retlw	.0	; 0

	 END



